/**************************************************************************
*  Copyright (c) 2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  24.03.2018  mifi  First Version.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <windows.h>
#include <stdio.h>
#include <time.h>

#include "stdint.h"
#include "mbedtls/config.h"
#include "mbedtls/pk.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define VERSION         "1.01"

#define BYTES_IN_LINE   16

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static char szInFile[_MAX_PATH];    /* Input file name */
static char szOutFile[_MAX_PATH];   /* Output file name */
static char szVariable[_MAX_PATH];  /* Variable name */

static char szOutFileHeader[_MAX_PATH];   /* Header file */
static char szOutFileSource[_MAX_PATH];   /* Source file */

static char szUpperName[_MAX_PATH];

static int                 nPrivateKey;
static mbedtls_pk_context   key;
static mbedtls_rsa_context *rsa;   

static const char HeaderStart[] = 
"/**************************************************************************\r\n"
"*  This file was created automatically by xpk2c.exe (www.emb4fun.de).\r\n"
"*\r\n"
"*  Source: %s\r\n"
"**************************************************************************/\r\n"
"#if !defined(__%s_H__)\r\n"
"#define __%s_H__\r\n"
"\r\n"
"/**************************************************************************\r\n"
"*  Extern Definitions\r\n"
"**************************************************************************/\r\n"
"\r\n";

static const char HeaderEnd[] =
"\r\n"
"#endif /* __%s_H__ */\r\n"
"/*** EOF ***/\r\n";

static const char SourceStart[] = 
"/**************************************************************************\r\n"
"*  This file was created automatically by xpk2c.exe (www.emb4fun.de).\r\n"
"*\r\n"
"*  Source: %s\r\n"
"**************************************************************************/\r\n"
"#define __%s_C__\r\n"
"\r\n"
"/**************************************************************************\r\n"
"*  Global Data\r\n"
"**************************************************************************/\r\n"
"\r\n";

//"const unsigned char %s[%d + %d] __attribute__((aligned(4))) =\r\n";

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  OutputStartMessage                                                   */
/*                                                                       */
/*  Output start message.                                                */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputStartMessage (void)
{
   printf("\n");   
   printf("xpk2c v%s compiled "__DATE__" "__TIME__"\n", VERSION);
   printf("(c) 2018 by Michael Fischer (www.emb4fun.de)\n");
   printf("\n");
} /* OutputStartMessage */

/*************************************************************************/
/*  OutputUsage                                                          */
/*                                                                       */
/*  Output "usage" message.                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputUsage (void)
{
  printf("Usage: xpk2c -i:Infile -o:Outfile -v:Variable\n");
  printf("\n");
  printf("  -i   Input file, e.g. -i:test.pem\n");
  printf("  -o   Output file without .c or .h, e.g. -o:test\n");
  printf("  -v   Variable, e.g. -v:_test_key\n");
} /* OutputUsage */

/*************************************************************************/
/*  WriteHeaderBlob                                                      */
/*                                                                       */
/*  In    : hOutFile, nLen, pName                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WriteHeaderBlob (FILE *hOutFile, int nLen, char *pName)
{
   /* extern const unsigned char <szVariable><pName>[nLen] __attribute__((aligned(4))); */
   
   fprintf(hOutFile, 
           "extern const unsigned char %s%s[%3d] __attribute__((aligned(4)));\r\n",
           szVariable, pName, nLen); 

} /* WriteHeaderBlob */

/*************************************************************************/
/*  WriteSourceBlob                                                      */
/*                                                                       */
/*  In    : hOutFile, nLen, pName                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void WriteSourceBlob (FILE *hOutFile, BYTE *pBuffer, DWORD dLen, char *pName)
{
   DWORD dIndex;
   DWORD dLineCount;
   DWORD dRestCount;
   DWORD dAddress;
   int    x;

   /* const unsigned char _rsa_private_1024b[888] __attribute__((aligned(4))) = */
   
   fprintf(hOutFile, 
           "const unsigned char %s%s[%d] __attribute__((aligned(4))) =\r\n",
           szVariable, pName, dLen); 
   fprintf(hOutFile, "{\r\n");
           
           
   dLineCount = dLen / BYTES_IN_LINE;
   dRestCount = dLen % BYTES_IN_LINE;

   dAddress = 0;
   if (dLineCount)
   {
      for(dIndex=0; dIndex<dLineCount; dIndex++)
      {
         fprintf(hOutFile, "   ");
         for(x=0; x<BYTES_IN_LINE; x++)
         {
            fprintf(hOutFile, "0x%02x", pBuffer[dAddress++]);
            if (dAddress != dLen)
            {
               fprintf(hOutFile, ", ");
            }
         }
         fprintf(hOutFile, "\r\n");
      }
   }
   
   if (dRestCount)
   {
      fprintf(hOutFile, "   ");
      for(dIndex=0; dIndex<dRestCount; dIndex++)
      {
         fprintf(hOutFile, "0x%02x", pBuffer[dAddress++]);
         if (dAddress != dLen)
         {
            fprintf(hOutFile, ", ");
         }
      }
      fprintf(hOutFile, "\r\n");
   }      
   
   fprintf(hOutFile, "};\r\n");
   fprintf(hOutFile, "\r\n");

} /* WriteSourceBlob */


/*************************************************************************/
/*  CreateOutputImage                                                    */
/*                                                                       */
/*  Create the output image file.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void CreateOutputImage (void)
{
   int        nLen;
   DWORD      dIndex;
   FILE      *hOutFile;
   BYTE        Buffer[1024];
   mbedtls_mpi N, P, Q, D, E, DP, DQ, QP;

   memset(Buffer, 0, sizeof(Buffer));

   mbedtls_mpi_init(&N);  mbedtls_mpi_init(&P); mbedtls_mpi_init(&Q);
   mbedtls_mpi_init(&D);  mbedtls_mpi_init(&E); mbedtls_mpi_init(&DP);
   mbedtls_mpi_init(&DQ); mbedtls_mpi_init(&QP);
   

   /* Create header and source filename */
   _snprintf(szOutFileHeader, sizeof(szOutFileHeader), "%s.h", szOutFile);
   _snprintf(szOutFileSource, sizeof(szOutFileSource), "%s.c", szOutFile);

   /* Create upper NAME for "#if !defined(__NAME__H__)" */
   _snprintf(szUpperName,  sizeof(szUpperName), "%s", szOutFile);
   for(dIndex=0; dIndex<strlen(szUpperName); dIndex++)
   {
      szUpperName[dIndex] = toupper(szUpperName[dIndex]);
   }

   
   /*
    * Create "header" file
    */
   hOutFile = fopen(szOutFileHeader, "wb");
   if (NULL == hOutFile)
   {
      /* File could not created */
      printf("Error: Could not create \"%s\"\n", szOutFileHeader);
      exit(1);
   }
   
   fprintf(hOutFile, HeaderStart, szInFile, szUpperName, szUpperName);

   if (0 == nPrivateKey)
   {
      /* Public Key */
      mbedtls_rsa_export(rsa, &N, NULL, NULL, NULL, &E);
      
      /* N */
      nLen = mbedtls_mpi_size(&N);
      WriteHeaderBlob(hOutFile, nLen, "_N");
      
      /* E */
      nLen = mbedtls_mpi_size(&E);
      WriteHeaderBlob(hOutFile, nLen, "_E");
   }
   else
   {
      /* Private Key */
      mbedtls_rsa_export(rsa, &N, &P, &Q, &D, &E);
      mbedtls_rsa_export_crt(rsa, &DP, &DQ, &QP);
      
      /* N */
      nLen = mbedtls_mpi_size(&N);
      WriteHeaderBlob(hOutFile, nLen, "_N ");
      
      /* P */
      nLen = mbedtls_mpi_size(&P);
      WriteHeaderBlob(hOutFile, nLen, "_P ");
      
      /* Q */
      nLen = mbedtls_mpi_size(&Q);
      WriteHeaderBlob(hOutFile, nLen, "_Q ");
      
      /* D */
      nLen = mbedtls_mpi_size(&D);
      WriteHeaderBlob(hOutFile, nLen, "_D ");

      /* E */
      nLen = mbedtls_mpi_size(&E);
      WriteHeaderBlob(hOutFile, nLen, "_E ");
      
      /* DP */
      nLen = mbedtls_mpi_size(&DP);
      WriteHeaderBlob(hOutFile, nLen, "_DP");
      
      /* DQ */
      nLen = mbedtls_mpi_size(&DQ);
      WriteHeaderBlob(hOutFile, nLen, "_DQ");
      
      /* QP */
      nLen = mbedtls_mpi_size(&QP);
      WriteHeaderBlob(hOutFile, nLen, "_QP");
   }
   
   fprintf(hOutFile, HeaderEnd, szUpperName);
   fclose(hOutFile);


   /*
    * Create "source" file
    */
   hOutFile = fopen(szOutFileSource, "wb");
   if (NULL == hOutFile)
   {
      /* File could not created */
      printf("Error: Could not create \"%s\"\n", szOutFileSource);
      exit(1);
   }
   
   fprintf(hOutFile, SourceStart, szInFile, szUpperName);
   
   if (0 == nPrivateKey)
   {
      /* Public Key */
      mbedtls_rsa_export(rsa, &N, NULL, NULL, NULL, &E);
      
      /* N */
      nLen = mbedtls_mpi_size(&N);
      mbedtls_mpi_write_binary(&N, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_N");
      
      /* E */
      nLen = mbedtls_mpi_size(&E);
      mbedtls_mpi_write_binary(&E, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_E");
   }
   else
   {
      /* Private Key */
      mbedtls_rsa_export(rsa, &N, &P, &Q, &D, &E);
      mbedtls_rsa_export_crt(rsa, &DP, &DQ, &QP);
      
      /* N */
      nLen = mbedtls_mpi_size(&N);
      mbedtls_mpi_write_binary(&N, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_N ");
      
      /* P */
      nLen = mbedtls_mpi_size(&P);
      mbedtls_mpi_write_binary(&P, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_P ");
      
      /* Q */
      nLen = mbedtls_mpi_size(&Q);
      mbedtls_mpi_write_binary(&Q, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_Q ");
      
      /* D */
      nLen = mbedtls_mpi_size(&D);
      mbedtls_mpi_write_binary(&D, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_D ");

      /* E */
      nLen = mbedtls_mpi_size(&E);
      mbedtls_mpi_write_binary(&E, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_E ");
      
      /* DP */
      nLen = mbedtls_mpi_size(&DP);
      mbedtls_mpi_write_binary(&DP, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_DP");
      
      /* DQ */
      nLen = mbedtls_mpi_size(&DQ);
      mbedtls_mpi_write_binary(&DQ, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_DQ");
      
      /* QP */
      nLen = mbedtls_mpi_size(&QP);
      mbedtls_mpi_write_binary(&QP, Buffer, nLen);
      WriteSourceBlob(hOutFile, Buffer, nLen, "_QP");
   }
   
   fprintf(hOutFile, "/*** EOF ***/\r\n");
   fclose(hOutFile);


   mbedtls_mpi_free(&N);  mbedtls_mpi_free(&P); mbedtls_mpi_free(&Q);
   mbedtls_mpi_free(&D);  mbedtls_mpi_free(&E); mbedtls_mpi_free(&DP);
   mbedtls_mpi_free(&DQ); mbedtls_mpi_free(&QP);

} /* CreateOutputImage */ 

/*************************************************************************/
/*  CheckValidKey                                                        */
/*                                                                       */
/*  Check for a valid key.                                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void CheckValidKey (void)
{
   int nRet;
   int nError = 1;
   
   nPrivateKey = 0;

   mbedtls_pk_init(&key);

   /* Check for private key */
   nRet = mbedtls_pk_parse_keyfile(&key, szInFile, NULL);
   if (0 == nRet)
   {
      nError      = 0;
      nPrivateKey = 1;
   }
   else
   {
      /* Check for public key */
      nRet = mbedtls_pk_parse_public_keyfile(&key, szInFile);
      if (0 == nRet)
      {
         nError      = 0;
         nPrivateKey = 0;
      }
   }
   
   if (1 == nError)
   {
      mbedtls_pk_free(&key);

      printf("Error: Input file <%s> is not a valid PEM key file\n", szInFile);
      exit(1);
   }
   else
   {
      if(MBEDTLS_PK_RSA == mbedtls_pk_get_type(&key))
      {
         rsa = mbedtls_pk_rsa(key);
      }
      else
      {
         mbedtls_pk_free(&key);

         printf("Error: Input file <%s> is not a valid RSA key file\n", szInFile);
         exit(1);
      }   
   }
   
} /* CheckValidKey */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : argc, argv                                                   */
/*  Out   : none                                                         */
/*  Return: 0 / error cause                                              */
/*************************************************************************/
int main (int argc, char **argv)
{
   int   nIndex;
   int   nLen;
   char *pString;
   
   szInFile[0]   = 0;
   szOutFile[0]  = 0;     
   szVariable[0] = 0;
   
   OutputStartMessage();
   
   if (argc != 4)
   {
      OutputUsage();
   }
   else
   {
      /* 
       * Parse command line 
       */
      for (nIndex=0; nIndex<argc; nIndex++)
      {
         /* Input file */
         if ((pString = strstr(argv[nIndex], "-i:")) != NULL)
         {
            pString += 3;  /* Jump over "-i:" */
            nLen     = strlen(pString);
            if (nLen < (sizeof(szInFile)-1))
            {
               strcpy(szInFile, pString);
            }
            else
            {
               printf("Error: Input name len > %d\n", (sizeof(szInFile)-1));
               exit(1);
            }
         } /* Parse input file name */
         
         /* Output file -o */
         if ((pString = strstr(argv[nIndex], "-o:")) != NULL)
         {
            pString += 3;  /* Jump over "-o:" */
            nLen     = strlen(pString);
            if (nLen < (sizeof(szOutFile)-1))
            {
               strcpy(szOutFile, pString);
            }
            else
            {
               printf("Error: Output name len > %d\n", (sizeof(szOutFile)-1));
               exit(1);
            }
         } /* Parse output file name */

         /* Variable name */
         if ((pString = strstr(argv[nIndex], "-v:")) != NULL)
         {
            pString += 3;  /* Jump over "-v:" */
            nLen     = strlen(pString);
            if (nLen < (sizeof(szVariable)-1))
            {
               strcpy(szVariable, pString);
            }
            else
            {
               printf("Error: Variable name len > %d\n", (sizeof(szVariable)-1));
               exit(1);
            }
         } /* Parse output file name */
         
      } /* end for loop */
      
      if (0 == szInFile[0])
      {
         printf("Error: No input file specified\n");
         exit(1);
      }
      
      if (0 == szOutFile[0])
      {
         printf("Error: No Output file specified\n");
         exit(1);
      }
      
      if (0 == szVariable[0])
      {
         printf("Error: No Variable specified\n");
         exit(1);
      }
      
      CheckValidKey();
      
      /*
       * Here the command line was parsed and no error was available
       */
      printf("Input file : %s\n", szInFile);
      printf("Output file: %s.c, %s.h\n", szOutFile, szOutFile);
      printf("Variable   : %s\n", szVariable);
      
      CreateOutputImage();
      
      mbedtls_pk_free(&key);
   }      
   
   return(0);
} /* main */

/*** EOF ***/
